from .clients.http_client import HttpClient
from .clients.ws_client import WSClient
from .cursor import Cursor
from .query_builder import QueryBuilder
from .config import config
from .utils import verify_table_and_id

class Connection:
    """
    This is the main object that users will interact with. 
    It is a wrapper around the client objects with additional functionality.
    
    Includes methods to emulate a pymysql connection object. This allows the use of libraries like pandas. 
    """
    connections = {}
    client: HttpClient = None
    _cursor = None
    def __init__(self, **kwargs):
        """
            Attempt to build a connection. If a client is passed, use that. 

            Args:
                client: Specify client or client type. Can be a string or a client object. If a string, it must be 'http' or 'ws'. If a client object, it must be an instance of HttpClient or WSClient.
                host: The host of the server. 
                port: The port of the server. 
                user: The user to login with. 
                password: The password to login with. 
                database: The database to use. 
                namespace: The namespace to use.
        """
        client = kwargs.pop('client', None)
        
        # Set client type based on config.
        Client = HttpClient if config.default_client.lower() in ['http', 'https'] else WSClient
        # If they specify a client type explicitly, use that instead.
        if isinstance(client, str):
            Client = HttpClient if client.lower() in ['http', 'https'] else WSClient

        # If they pass a client object, use that. Otherwise, create a new client.
        if isinstance(client, HttpClient) or isinstance(client, WSClient):
            self.client = client
        elif kwargs:
            self.client = Client(**kwargs)
        else:
            self.client = Client()


    def cursor(self) -> Cursor:
        if not self._cursor:
            self._cursor = Cursor(self.client)
        return self._cursor

    def commit(self):
        pass

    # allow entry into query builder.
    def table(self, table) -> QueryBuilder:
        return QueryBuilder(self.client).table(table)

    def relate(self, noun1, verb, noun2, data=None):
        """Create a relationship between two records."""
        return QueryBuilder(self.client).relate(noun1, verb, noun2, data)

    def upsert(self, table, data=None, keys=['id']):
        """Update or create a record in the specified table"""
        if not data:
            data = table
            table, id = verify_table_and_id(None, data.get('id'))

        if ':' in table:
            if keys != ['id']:
                raise ValueError("Cannot upsert a record with a key that is not 'id' when using the table:id syntax.")
            table, id = verify_table_and_id(table, data.get('id'))
            r = self.get(table, id)
            if r:
                return self.update(table, data)
            else:
                return self.create(table, data)
            
        if isinstance(keys, str):
            keys = [keys]

        for k in keys:
            if k not in data:
                raise ValueError(f"Cannot upsert a record without a key. Key {k} not found in data.")

        conditions = []
        for key in keys:
            if key == 'id':
                table, id = verify_table_and_id(table, data.get('id'))
                conditions.append([key, f"{table}:{id}"])
            else:
                conditions.append([key, data[key]])
        existing = QueryBuilder(self).table(table).where(conditions).exists()
        if existing:
            return self.update(table, data)
        else:
            return self.create(table, data)

    # These methods are just wrappers around the client methods. They are here for autocomplete. If anyone knows a better way to do this, please let me know.
    def select(self, sql):
        return self.client.select(sql)

    def query(self, sql):
        return self.client.query(sql)

    def get(self, table, id=None):
        return self.client.get(table, id)

    def insert(self, table, data):
        return self.client.insert(table, data)

    def create(self, table, data):
        return self.client.create(table, data)

    def update(self, table, data=None):
        return self.client.update(table, data)

    def delete(self, table, id=None):
        return self.client.delete(table, id)

    def drop(self, table):
        return self.client.drop(table)

    # any other methods should just be passed to the client
    def __getattr__(self, name):
        return getattr(self.client, name)

    
    


    

    
    



from .connections import Connection
from .query_builder import QueryBuilder


class Model:
    """
    A class to represent a model that can be used to load and store data from the database.

    This is not fully implemented. Do not use. 
    """
    _table = None
    _data = {}
    _client = None
    _id = None

    def __init__(self):
        if not self._table:
            raise Exception('Table name not set.')
        
        self._connection = Connection()

    @property
    def _query_builder(self):
        return QueryBuilder(self._connection)

    def get(self):
        """
        Get a single row from the database.
        """
        res = self._query_builder.table(self._table).first()
        if res:
            self._data = res
            self._id = res['id']
        return self
        

    def all(self):
        """
        Get all rows from the database.
        """
        res = self._query_builder.table(self._table).get()
        all = []
        for row in res:
            model = self.__class__()
            model._data = row
            model._id = row['id']
            all.append(model)
        return all

    def find(self, id):
        """
        Find a row by id.
        """
        res = self._query_builder.table(self._table).find(id)
        if res:
            self._data = res
            self._id = res['id']
        return self

    def where(self, *args):
        """
        Find a row by column and value.

        TODO: This should return a new model instance. 
        """
        return self._query_builder.table(self._table).where(*args)

    def save(self):
        """
        Save the model to the database.
        """
        if self._id:
            self._data['id'] = self._id
            return self._connection.update(self._table, self._data)
        else:
            return self._connection.insert(self._table, self._data)



class QueryBuilder:
    """
    This class is used to build a query to send to the database. 
    Similar to Laravel and Orator's query builder.
    """
    client = None
    _where = []
    _limit = None
    _order_by = None
    _group_by = None
    _select = None
    _table = None
    _type = 'select'
    _relate = None
    _fetch = []
    _data = None
    _escape_table = None

    def __init__(self, client):
        self.client = client
        self._where = []
        self._limit = None
        self._order_by = None
        self._group_by = None
        self._select = None
        self._table = None
        self._type = 'select'
        self._relate = None
        self._fetch = []
        self._data = None
        self._escape_table = None

    def where(self, *args) -> 'QueryBuilder':
        """
            Add a where clause to the query.

            You can pass an array of where clauses, or a series of arguments.
            If you pass an array, each element should be in the format ['column', 'operator', 'value'].
            If you pass a series of strings, they should be in the format 'column', 'operator' ,'value'. If no operator is specified, '=' is used.
        """
        self._where.append(args)

        return self

    def where_in(self, column, values) -> 'QueryBuilder':
        """
        Add a where in clause to the query. 
        """
        self._where.append([values, 'CONTAINS', column])
        return self

    def where_contains(self, column, value) -> 'QueryBuilder':
        return self.where_in(column, value)

    def where_not_in(self, column, values) -> 'QueryBuilder':
        """
        Add a where not in clause to the query. 
        """
        self._where.append([values, 'CONTAINSNOT', column])
        return self

    def where_contains_not(self, column, value) -> 'QueryBuilder':
        return self.where_not_in(column, value)

    def or_where(self, *args) -> 'QueryBuilder':
        """
        Add an or where clause to the query.
        """
        self._where.append(['OR'])
        return self.where(*args)

    def select(self, *args) -> 'QueryBuilder':
        """
        Set the columns to select.
        """
        self._select = args
        return self

    def insert(self, data):
        """
        Insert one or many rows.
        """
        return self.client.insert(self._table, data)

    def update(self, data):
        """
        Update fields.
        """
        self._type = 'update'
        self._data = data
        # we use the query method because the update api only supports 1 row at a time.
        return self.client.query(self._build_query())

    def relate(self, noun1, verb, noun2, data=None):
        """
        Relate 2 objects. Noun->Verb->Noun. You can also set data to be stored with the relationship.
        """
        self._type = 'relate'
        self._relate = [noun1, verb, noun2]
        self._data = data
        return self.client.query(self._build_query())

    def table(self, table) -> 'QueryBuilder':
        """
        Set the table to query.
        """
        self._table = table
        return self

    def limit(self, limit) -> 'QueryBuilder':
        """
        Set the limit.
        """
        self._limit = limit
        return self

    def order_by(self, column, direction='ASC') -> 'QueryBuilder':
        """
        Set the order by.
        """
        self._order_by = [column, direction]
        return self

    def group_by(self, column) -> 'QueryBuilder':
        """
        Set the group by.
        """
        self._group_by = column
        return self

    def get(self):
        """
        Execute the query and return the result.
        """
        return self.client.query(self._build_query())

    def first(self):
        """
        Execute the query and return the first result.
        Return None if no results are found.
        """
        self._limit = 1
        results = self.get()
        if results:
            return results[0]
        return None

    def exists(self):
        """
        Execute the query and return True if any results are found.
        """
        self._select = ['id']
        self._limit = 1
        return bool(self.get())

    def fetch(self, column):
        """
        Add a fetch clause to the query.
        """
        # there is currently a bug that fetch sometimes fails without an order by. Add one in if not there.
        if not self._order_by:
            self._order_by = ['id', 'ASC']
        self._fetch.append(column)
        return self

    def count(self):
        """
        Execute the query and return the number of results.
        """
        self._select = ['count()']
        if not self._group_by:
            self._group_by = 'all'
        return self.get()[0]['count']

    def sum(self, column):
        """
        Execute the query and return the sum of a column.
        """
        self._select = [f'math::sum({column})']
        if not self._group_by:
            self._group_by = 'all'
        return self.get()[0][f'math::sum']

    def avg(self, column):
        """
        Execute the query and return the average of a column.
        """
        self._select = [f'math::mean({column})']
        if not self._group_by:
            self._group_by = 'all'
        return self.get()[0][f'math::mean']

    def mean(self, column):
        return self.avg(column)

    def max(self, column):
        """
        Execute the query and return the max of a column.
        """
        self._select = [f'math::max({column})']
        if not self._group_by:
            self._group_by = 'all'
        return self.get()[0][f'math::max']

    def min(self, column):
        """
        Execute the query and return the min of a column.
        """
        self._select = [f'math::min({column})']
        if not self._group_by:
            self._group_by = 'all'
        return self.get()[0][f'math::min']

    def to_sql(self):
        """
        Return the query as a string.
        """
        return self._build_query()

    def _quote(self, value):
        """
        Quote a value for use in a query.
        """
        if isinstance(value, str):
            return f"'{self._escape_string(value)}'"
        return str(value)

    def _escape_string(self, value):
        """escapes *value* without adding quote.
        Value should be unicode
        """
        escape_table = self._build_escape_table()        
        return value.translate(escape_table)

    def _build_escape_table(self):
        if not self._escape_table:
            self._escape_table = [chr(x) for x in range(128)]
            self._escape_table[0] = "\\0"
            self._escape_table[ord("\\")] = "\\\\"
            self._escape_table[ord("\n")] = "\\n"
            self._escape_table[ord("\r")] = "\\r"
            self._escape_table[ord("\032")] = "\\Z"
            self._escape_table[ord('"')] = '\\"'
            self._escape_table[ord("'")] = "\\'"

        return self._escape_table

    def _build_query(self):
        """
        Build the query.
        """
        if self._type == 'select':
            return self._build_select()
        elif self._type == 'update':
            return self._build_update()
        elif self._type == 'relate':
            return self._build_relate()

    def _build_select(self):
        """
        Build the query.
        """
        query = 'SELECT '
        if self._select:
            query += ', '.join(self._select)
        else:
            query += '*'
        query += f' FROM {self._table}'
        if self._where:
            query += ' WHERE '
            query += self._build_where(self._where)
        if self._group_by:
            query += f' GROUP BY {self._group_by}'
        if self._order_by:
            query += f' ORDER BY {self._order_by[0]} {self._order_by[1]}'
        if self._limit:
            query += f' LIMIT {self._limit}'
        if self._fetch:
            query += ' FETCH '
            query += ', '.join(self._fetch)

        return query

    def _build_where(self, _where):
        """
        Build the where clause.
        """
        query = ''
        for where in _where:
            # Check for an OR statement
            if len(where) == 1 and isinstance(where[0], str):
                query = query[:-4] # remove the previous 'AND'
                query += f' {where[0]} '
                continue
            # if all where elements are lists, it's a nested where
            if all(isinstance(x, list) for x in where):
                query += '('
                query += self._build_where(where)
                query += ')'
                query += ' AND '
                continue

            # if we only have 2 arguments, assume equals
            if len(where) == 2:
                where = [where[0], '=', where[1]]
            
            # if the first position is a list, its probably a contains clause. They are in reverse order in surreal. 
            if isinstance(where[0], list):
                query += '['
                for value in where[0]:
                    if isinstance(value, str):
                        query += f'{self._quote(value)}, '
                    else:
                        query += f'{value}, '
                query = query[:-2]
                query += f'] {where[1]} {where[2]}'
            elif isinstance(where[2], str):
                query += f'{where[0]} {where[1]} {self._quote(where[2])}'
            else:
                query += f'{where[0]} {where[1]} {where[2]}'
            query += ' AND '
            
        query = query[:-5]   
        return query     

    def _build_update(self):
        """
        Build the query.
        """
        query = f'UPDATE {self._table} SET '
        for key, value in self._data.items():
            if isinstance(value, str):
                query += f'{key}={self._quote(value)}, '
            else:
                query += f'{key}={value}, '
        query = query[:-2]
        if self._where:
            query += ' WHERE '
            query += self._build_where(self._where)         
        return query

    def _build_relate(self):
        """
        Build the query.
        """
        query = f'RELATE {self._relate[0]}->{self._relate[1]}->{self._relate[2]} '
        if self._data:
            query += ' SET '
            for key, value in self._data.items():
                if isinstance(value, str):
                    query += f'{key}={self._quote(value)}, '
                else:
                    query += f'{key}={value}, '
            query = query[:-2]

        return query




def verify_table_and_id(table=None, id=None):
    """
    Verifies the table and id parameters and returns them seperately.

    Args:
        table (str): May take the form of 'table' or 'table:id'. 
        id (str): May take the form of 'id' or 'table:id'.

    Returns:
        table (str): The table name.
        id (str): The id.

    Raises:
        ValueError: If the table and id conflict.
    """

    table_table, table_id = None, None
    if table is not None:
        if ':' in table:
            table_table, table_id = table.split(':')
        else:
            table_table = table

    id_table, id_id = None, None
    if id is not None:
        if ':' in id:
            id_table, id_id = id.split(':')
        else:
            id_id = id

    if table_table is not None and id_table is not None and table_table != id_table:
        raise ValueError("table from ID doesn't match the given table", table_table, id_table)
    if table_id is not None and id_id is not None and table_id != id_id:
        raise ValueError("id from table doesn't match the given id", table_id, id_id)

    table = table_table or id_table
    id = table_id or id_id

    return table, id






from .clients.http_client import HttpClient


class Cursor:
    """
    This is to simulate a pymysql Cursor object. This allows the use of libraries like pandas that expect a cursor object.

    This is instantiated by the connection object. You should not need to instantiate this directly.
    """
    def __init__(self, client: HttpClient):
        self.client = client

    def execute(self, sql):
        self.data = self.client.query(sql)

    @property
    def description(self):
        if self.data:
            keys = self.data[0].keys()
            return [(key, None, None, None, None, None, None) for key in keys]
        return []

    def fetchall(self):
        return self.data

    def close(self):
        self.client.close()


# add my own installation to path. It is 2 levels up from this file.
import sys, os

sys.path.insert(0, (os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))))
# sys.path.insert(0, 'C:/Users/Mike/Sync/code/mypysurreal/')


import pysurrealdb as surreal

# conn = surreal.connect(user='test', password='test', database='test', namespace='test')
conn = surreal.connect(user='test', password='test', database='main', namespace='main', client='ws')
# conn = surreal.connect(host='https://sparkling-cherry-9759.fly.dev', user='proth', password='passicus', database='main', namespace='main')



# conn.drop('person')
# conn.insert('test', {'id': 1 , 'name': 'test', 'age': 20.65, 'description': 'Test'})
# conn.query("insert into test {'id': 2 , 'name': 'test2', 'age': 20.65}")
# conn.insert('person', {'id': 2 , 'name': 'test', 'age': 2, 'description': '"This" is `a` test of strange encoding characters.'})
# conn.insert('test', {'id': 3 , 'name': 'test', 'age': 12, 'description': '"This" is `a` test of strange encoding characters.'})
# conn.insert('person', {'id':69 ,'name': "'test'", 'age': 42, 'description': "'This' is `a` test of strange encoding characters."})

# print(conn.query('select total from profiler'))
# print(conn.get("profiler"))
# conn.drop('profiler')
# conn.insert('documents', {'filename': '📜Writing Week 15 - Lesson 15📜.docx'})

conn.drop('test')
conn.create('test:test2', {'id':'test2', 'name': 'test1'})
# conn.insert('test', {'id': 'test', 'name': 'test1'})
# r = conn.upsert('test', {'id': 'test', 'name': 'test2'})
r = conn.get('test')
# r = conn.get('test', 'testicles')
# r = conn.query("SELECT id FROM test WHERE ((id = 'test:test')) LIMIT 1")
# r = conn.table('test').insert([{ 'id': 'test', 'name': 'test' }, { 'id': 'test2', 'name': 'test2' }])
print(r)

# Sending b'{"filename": "\\ud83d\\udcdcWriting Week 15 - Lesson 15\\ud83d\\udcdc.docx"}' to http://localhost:8000/key/documents
# Sending b'select * from documents where filename = "\xf0\x9f\x93\x9cWriting Week 15 - Lesson 15\xf0\x9f\x93\x9c.docx"' to http://localhost:8000/sql


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysurrealdb', 'pysurrealdb.clients', 'pysurrealdb.tests']

package_data = \
{'': ['*']}

install_requires = \
['requests']

setup_kwargs = {
    'name': 'pysurrealdb',
    'version': '0.3.0',
    'description': 'A library to connect to SurrealDB.',
    'long_description': '# PySurrealDB\n\n[![Join the chat at https://gitter.im/pysurrealdb/community](https://badges.gitter.im/pysurrealdb/community.svg)](https://gitter.im/pysurrealdb/community?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n\nAn unofficial library to connect to SurrealDB.\n\nMinimal dependencies, easy to use.\n\n---\n## Getting Started\n\nIf you don\'t already have it, install [SurrealDB](https://surrealdb.com/docs/start/installation)\n\nLinux: \n``` bash\n$ curl -sSf https://install.surrealdb.com | sh\n# then\n$ surreal start --user test --pass test\n```\n\n\n### Install PySurrealDB\n```\npip install pysurrealdb\n```\n\n\n### Examples\n\n```python\nimport pysurrealdb as surreal\n\nconn = surreal.connect(user=\'test\', password=\'test\')\n\nconn.create(\'person\', {\'name\': \'Mike\'})\nconn.query(\'select * from person\')\n```\n\nYou can specify additional connection info either in the connect call, or in a config file.\n\n```python\nimport pysurrealdb as surreal\nconn = surreal.connect(host=\'surreal.com\', port=8000, user=\'user\', password=\'pass\', database=\'db\', namespace=\'ns\')\n```\n\nBoth http and websocket are supported. Specify which to use with the client keyword.\n```python\nconn = surreal.connect(client=\'websocket\')\n# Websocket was added and made the default as of version 0.3. Try http if you run into issues, and please report any bugs you find!\n```\n\nOptional Config file:\n```python\n# use a configured connection. \nconn = surreal.connection(\'default\')\n# Requires pysurrealdb.json file. Place it in your root directory, or specify the file location with the env variable \'PYSURREALDB_CONFIG\'.\n\nExample pysurrealdb.json:\n{\n    "connections": {\n        "default": {\n            "host": "localhost",\n            "port": 8000,\n            "user": "test",\n            "password": "test"\n            "database": "test",\n            "namespace": "test",\n            "client": "http",\n        }\n    }\n}\n\n# when using a config file, you do not even need to connect, you can access most functions directly:\nimport pysurrealdb as surreal\n\nsurreal.query(\'select * from test\') # uses the last connection from connect() or the default connection if connect() has not been called.\n```\n\n\n## Query Builder\n\nYou can write queries using Laravel and Orator style syntax:\n```python\nimport pysurrealdb as surreal\nconn = surreal.connection()\n\n# setup data\nconn.drop(\'person\')\nconn.insert(\'person\', [{\'name\': \'Mike\', \'age\': 31}, {\'name\':\'Mr P\'}])\n\n# query builder examples\nfirst_person = conn.table(\'person\').where(\'name\', \'Mike\').first()\n\nadults = conn.table(\'person\').where(\'age\', \'>=\', 18).order_by(\'age\', \'desc\').limit(10).get()\n```\n\n## Methods\nSome of the basic methods available:\n```python\nquery(sql)\nget(table, id=\'\')\ninsert(table, data)\ncreate(table, data)\nupdate(table, data)\nupsert(table, data)\ndelete(table, id)\ndrop(table)\nrelate(noun, verb, noun2, data={})\n\n# Most methods accept a table or table:id as the main arguement. The data is also checked for an ID when relevant.\n```\n\n\nThis project is a work in progress. Questions and feedback are welcome! Please create an issue or use the gitter chat link at the top. Thanks!\n',
    'author': 'Aurelion314',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import requests

from .clients.http_client import HttpClient
from .connections import Connection
from .model import Model
from .query_builder import QueryBuilder
from .config import config

Client = Instance = Connection # I've seen some people prefer to use the name Client/Instance instead of Connection. This is just an alias. Not to be confused with the HttpClient class, which is the interface to the surreal api.

def connect(host=None, port=None, user=None, password=None, database=None, namespace=None, client=None) -> Connection:
    """ 
    Connect to the SurrealDB server. 

    Args:
        client: Specify client or client type. Client type can be 'http' or 'websocket'. If passing a client object, it must be an instance of HttpClient or WSClient.
        host: The host of the server. defaults to localhost.
        port: The port of the server. defaults to 8000.
        user: The user to login with.
        password: The password to login with.
        database: The database to use. defaults to 'main'.
        namespace: The namespace to use. defaults to 'main'.
    
    This function caches the current connection. 
    """

    if isinstance(host, dict):
        # if host is a dict, assume it's a config object
        Connection.connections['current'] = Connection(**host)
    else:
        Connection.connections['current'] = Connection(host=host, port=port, user=user, password=password, database=database, namespace=namespace, client=client)
    return Connection.connections['current']

def connection(name:str = None) -> Connection:
    """
    Connect to the SurrealDB server. Similar to connect(), but uses a connection name from the config file.

    Args:
        name: The name of the connection. This is used to look up the connection in the config file. If not specified, the most recent connection is returned. If no current connection exists, a default connection is attempted.
    """
    conn = None
    # If no name is specified, return the most recent connection
    if name is None and 'current' in Connection.connections:
        conn = Connection.connections['current']

    elif name:
        # have we already created a connection with this name?
        if name in Connection.connections:
            conn = Connection.connections[name]
        elif name in config.connections: 
            # if not, create a new connection
            conn = Connection(**config.connections[name])
            Connection.connections[name] = conn
        else:
            raise Exception(f'SurrealDB: Connection "{name}" not found.')

    # check for a default config
    elif config.connections.get('default'):
        conn = Connection(**config.connections['default'])
        Connection.connections['default'] = conn

    if not conn:
        # try to connect to localhost
        conn = Connection()

    # cache the connection
    Connection.connections['current'] = conn
    
    return conn


## These are utility functions to allow the use of query methods without having to create a connection object. Note: They only work if a connection has already been created, or a default connection is available.
def table(name: str) -> QueryBuilder:
    # Return a query builder for the specified table.
    return QueryBuilder(connection()).table(name)

def select(*args, **kwargs):
    return connection().select(*args, **kwargs)

def query(*args, **kwargs):
    return connection().query(*args, **kwargs)

def get(*args, **kwargs):
    return connection().get(*args, **kwargs)

def insert(*args, **kwargs):
    return connection().insert(*args, **kwargs)

def create(*args, **kwargs):
    return connection().create(*args, **kwargs)

def update(*args, **kwargs):
    return connection().update(*args, **kwargs)

def upsert(*args, **kwargs):
    return connection().upsert(*args, **kwargs)

def delete(*args, **kwargs):
    return connection().delete(*args, **kwargs)

def drop(*args, **kwargs):
    return connection().drop(*args, **kwargs)

def relate(*args, **kwargs):
    return connection().relate(*args, **kwargs)


import requests, json, sys
from requests.auth import HTTPBasicAuth
from ..query_builder import QueryBuilder
from ..config import config

class HttpClient:
    """
    Representation of a connection with a SurrealDB server
    
    You should not need to use this class directly. Instead, use the Connection class via the connect() method.
    """
    request_size_limit = 14000
    def __init__(self, host=None, port=None, user=None, password=None, database=None, namespace=None):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.namespace = namespace
        self.session = requests.Session()
        self.auth = HTTPBasicAuth(user, password)

        if host:
            if 'http:' not in self.host and 'https:' not in self.host:
                self.host = 'http://' + self.host
        else:
            if config.warnings: print("SurrealDB: No host specified. Using localhost.")
            self.host = 'http://localhost'
        if not port:
            if host and 'https:' in host:
                self.port = 443
            else:
                self.port = 8000
            if config.warnings: print("SurrealDB: No port specified. Using", self.port)
        if not namespace:
            if config.warnings: print("SurrealDB: No namespace specified. Using main.")
            self.namespace = 'main'
        if not database:
            if config.warnings: print("SurrealDB: No database specified. Using main.")
            self.database = 'main'

        self._set_headers()

    def _set_headers(self, headers=None):
        if not headers:
            headers = {
                'Content-Type': 'application/json',
                'Accept':'application/json',
                'ns': self.namespace,
                'db': self.database,
            }

        self.session.headers.update(headers)
        
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def _send(self, data, method='POST', endpoint='sql'):
        """Send a request to SurrealDB and return the response."""
        url = f"{self.host}:{self.port}/{endpoint}"

        if not isinstance(data, str):
            data = json.dumps(data, default=str, ensure_ascii=False)
        
        data = data.encode('utf-8')
        response = self.session.request(method, url, data=data, auth=self.auth)
        
        if not response.ok:
            raise ConnectionError("Request to SurrealDB failed.", response.content)
        
        r = response.json()
        if len(r) > 1:
            results = []
            for row in r:
                if row['status'] != 'OK':
                    raise ConnectionError("Query failed.", row['result'])
            results.append(row['result'])
            return results
                
        if r[0]['status'] != 'OK':
            raise ConnectionError("Query failed.", r[0])
        return r[0]['result']

    def _send_chunks(self, data, method='POST', endpoint='sql'):
        """Send a request to SurrealDB and return the response."""
        # Note: Not currently used, but may be useful in the future
        url = f"{self.host}:{self.port}/{endpoint}"
        if isinstance(data, str):
            r = self.session.request(method, url, data=data, auth=self.auth, stream=True, timeout=5)
        else:
            r = self.session.request(method, url, json=data, auth=self.auth, stream=True, timeout=5)
        r.raise_for_status()

        receive_timeout = 5
        import time
        content = ''
        start = time.time()
        for chunk in r.iter_content(1024):
            if time.time() - start > receive_timeout:
                raise ConnectionError('SurrealDB timeout reached')

            if chunk:
                content += chunk.decode('utf-8')
        
        content = json.loads(content)
        return content[0]['result']

    def close(self):
        """Close the connection to SurrealDB."""
        self.session.close()

    def select_db(self, database):
        """Select a database."""
        self.database = database
        self._set_headers()

    def select_namespace(self, namespace):
        """Select a namespace."""
        self.namespace = namespace
        self._set_headers()

    def use(self, ns, db):
        """Select a namespace and database."""
        self.select_namespace(ns)
        self.select_db(db)

    def query(self, sql):
        """Execute an SQL query and return the result."""
        return self._send(sql)

    def select(self, sql):
        """Execute an SQL query and return the result."""
        return self.query(sql)

    def create(self, table, data):
        """Create one or many records in a SurrealDB table."""
        if isinstance(data, list):
            return self.create_many(table, data)
        return self.create_one(table, data)

    def insert(self, table, data):
        """Insert one or many records in a SurrealDB table."""
        return self.create(table, data)

    def create_large(self, table, data):
        """Create a record in a SurrealDB table that is larger than Surreal's request limit."""
        # we have to use query method for this, since it has a larger limit than they key endpoints.
        # format needs to be like this: INSERT INTO company {name: 'SurrealDB',founded: "2021-09-10",founders: [person:tobie, person:jaime], tags: ['big data', 'database']}
        data_string = ','.join([f"{k}: {json.dumps(v, default=str)}" for k, v in data.items()])
        return self.query(f"INSERT INTO {table} " + "{"+data_string +"}")

    def create_one(self, table, data=None):
        """Create a new record in a SurrealDB table."""
        # if the size of data is over 16kib, we must use create_large
        if self.getsizeof(data) > self.request_size_limit: # use 14000 to be safe
            return self.create_large(table, data)

        if 'id' in data:
            id = str(data['id'])
            if ':' in id:
                id_table, id = id.split(':')
                if id_table != table:
                    raise ValueError("Cannot create a record with an ID that doesn't match the table.")
            return self._send(data, method='POST', endpoint=f'key/{table}/{id}')

        return self._send(data, method='POST', endpoint=f'key/{table}')

    def create_many(self, table, data=None):
        """Create a new record in a SurrealDB table."""
        #TODO: Currently the SurrealDB API doesn't support multiple records, so we loop through and create them one at a time. However if the API is updated we should update this too for efficiency.
        results = []
        for row in data:
            r = self.create_one(table, row)
            results.append(r)
        return results

    def update(self, table, data=None):
        """Update a record in a SurrealDB table."""
        # we need an ID to update. It could be in either the table or data argument.
        id = None
        if ':' in table:
            table, id = table.split(':')

        if 'id' not in data and not id:
            raise ValueError("Cannot update a record without an ID.")
        
        id = str(data['id']) if id is None else id
        if ':' in id:
            id_table, id = id.split(':')
            if id_table != table:
                raise ValueError("Cannot update a record with an ID that doesn't match the table.", id_table, table)

        # check if the size of data is over 16kib, if so we must use update_large
        if self.getsizeof(data) > self.request_size_limit: # use 14000 to be safe
            return self.update_large(table, id, data)

        return self._send(data, method='PUT', endpoint=f'key/{table}/{id}')

    def update_large(self, table, id, data):
        """Update a record containing over 16kb of data."""
        # we have to use query method for this, since the surreal key endpoints all have a 16kb limit
        # format needs to be like this: UPDATE company SET name = 'SurrealDB',founded = "2021-09-10",founders = [person:tobie, person:jaime], tags = ['big data', 'database'] WHERE id = company:1
        data_string = ','.join([f"{k} = {json.dumps(v, default=str)}" for k, v in data.items()])
        return self.query(f"UPDATE {table} SET " + data_string + f" WHERE id = {table}:{id}")

    def upsert(self, table:str, data:dict, key=['id']):
        raise DeprecationWarning("upsert from the client is deprecated. Please use connection object instead.")

    def delete(self, table, id=None):
        """Delete a record in a SurrealDB table."""
        if ':' in table:
            table, id = table.split(':')
        if not id:
            raise ValueError("Cannot delete a record without an ID. If you meant to delete the entire table, use the drop() method.")
        return self._send(None, method='DELETE', endpoint=f'key/{table}/{id}')

    def drop(self, table):
        """Drop a SurrealDB table."""
        return self._send(None, method='DELETE', endpoint=f'key/{table}')

    def get(self, table, id=None):
        """Get a record from a SurrealDB table."""
        if ':' in table:
            table, id = table.split(':')
        if not id:
            return self._send(None, method='GET', endpoint=f'key/{table}')
        results = self._send(None, method='GET', endpoint=f'key/{table}/{id}')
        return results[0] if results else None

    def getsizeof(self, data):
        """Get the size of a dictionary in bytes."""
        size = 0

        if isinstance(data, list):
            for item in data:
                size += self.getsizeof(item)
            return size
        if isinstance(data, dict):
            for k, v in data.items():
                size += self.getsizeof(k)
                size += self.getsizeof(v)
            return size
        
        return sys.getsizeof(data)
            